#!BPY

"""
Name: 'The Movies head files (.hd)'
Blender: 3.3.0
Group: 'Import'
Tooltip: 'Import from Lionhead Studios The Movies actor Head file format. (.hd)'
"""

__author__ = ["Glen Rickey"]
__url__ = ("Director's Cut Modding Foundry","http://www.dcmodding.com")
__version__ = "1.01 10-25-2009"
__bpydoc__ = """\

hd Importer

This script imports an hd file from Lionhead Studio Limited's The Movies
into blender for editing. It is not supported by either Lionhead or
Activision.

Known Problems:

You will have to start with the generic_facialskin.msh

This version converted to run in Blender 3.3.0 by
Cary Semar 9/17/2022


"""

# ***** BEGIN GPL LICENSE BLOCK *****
#
# Script copyright (C) Mark S Andrews, Nick Hudson, Glen Rickey
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# ***** END GPL LICENCE BLOCK *****
# --------------------------------------------------------------------------


# Import Modules
#import Blender
#from Blender import Image, Material, Texture, Armature, Window
#from Blender import Mesh, NMesh, Group
#import struct, os, math
import bpy, os, time, struct
from bpy.types import Operator
from bpy_extras.io_utils import ImportHelper
from bpy.props import StringProperty
from time import time

hh_CURRENT_VERSION = 4

#***********************************************
# helper classes & functions for files
#***********************************************

def read_item(struct_type, stream):
    value = struct.unpack(struct_type,
        stream.read(struct.calcsize(struct_type)))
    if len(value) == 1:
        value = value[0]
    return value

def read_nts(stream, length=32):
    s = stream.read(length)
    return s.decode("ASCII")

def pad(stream):
    align = stream.tell()
    if ((align % 4)>0): 
        #skip padding to next 4-byte boundary
        stream.read(4 - (align % 4))

class decoder(object):
    _struct = ''
    _fields = {}
    _flags = {}
    
    def __init__(self, stream):
        self._value = struct.unpack(self._struct,
            stream.read(struct.calcsize(self._struct)))
    
    def __getattr__(self, item):
        if item in self._fields.keys():
            keyid = self._fields[item]
            return self._value[keyid]
        elif item in self._flags.keys():
            flag, bit = self._flags[item]
            return bool(self._value[flag] & bit)
        else:
            raise AttributeError(item)

    def __repr__(self):
        return repr(self._value)

def get_image_path(basepath, filename):
    p = os.path.join(basepath, filename)
    if os.path.isfile(p):
        return p
    else:
        dirs = ["people","props","costumes","accessories",
            "hair","makeup","lightmap","backdrops"]
        for dir in dirs:
            p = os.path.join(basepath, dir, filename)
            if os.path.isfile(p):
                return p
    
#***********************************************
# class definitions for hd file
#***********************************************   

class hd_header(decoder):
    _struct = "BBBBLL"
    _fields = {"magic":0,"unk1":4, "unk2":5}
    _flags = {}
    def __init__(self, stream):
        decoder.__init__(self, stream)
        self.texture1 = read_nts(stream,32)
        self.texture2 = read_nts(stream,32)
        self.headfile1 = read_nts(stream, 32)
        self.headfile2 = read_nts(stream, 32)
        self.mix = read_item("fxxxx", stream)
        
class facevert(object):
    def __init__(self, stream):
        self.xpos = read_item("f", stream)
        self.ypos = read_item("f", stream)
        self.zpos = read_item("f", stream)
        self.xnor = read_item("f", stream)
        self.ynor = read_item("f", stream)
        self.znor = read_item("f", stream)
        
class faceverts(list):
    def __init__(self, stream):
        verts = []
        for v in range(996): #996 is the number of verts in the TM generic facial mesh
            vert = verts.append(facevert(stream))
        super(faceverts, self).__init__(verts)

class param(decoder):
    _struct = "ffffff"
    _fields = {"p1":0, "p2":1, "p3":2, "p4":3, "p5":4, "p6":5 }
    _flags = {}
    
class params(list):
    def __init__(self, stream):
        ps = []
        for p in range(996): #996 is the number of verts in the TM generic facial mesh
            par = ps.append(param(stream))
        super(params, self).__init__(ps)
        
#***********************************************
        
def load_hd (filename):
    start_time = time()
    prev_time = start_time
#    in_editmode = Window.EditMode()
#    if in_editmode:
#        Window.EditMode(0)
    stream = open(filename,"rb")
    filepath = os.path.dirname(filename)
    filebase = os.path.basename(filename)
    fullbase = os.path.splitext(filebase)[0]
    basename = fullbase[0:21] #Blender has a $%$@ 21-char limit!
    basedir = os.path.split(filepath)[0]
    imagedir = os.path.join(basedir,"textures")
    
    header = hd_header(stream)
    if header.magic != 4 :
        print('File is not a TM Head file')
    print('header processed: %.4f sec.' % (time()-prev_time))
    prev_time = time()
    
 #   scn = bpy.data.scenes.active
    scn = bpy.context.scene
    dirs = bpy.utils.script_paths()
    datadir = dirs[1]+'\\bpydata'
    print('datadir = ',datadir)
#    datadir = Blender.Get("datadir")
    #scene.setLayers([1])
    bpath = datadir + '\\generic_facial_skin.blend'
    names = ['00.00','1094395392','Object.000','Object.001','Object.002']
    with bpy.data.libraries.load(bpath,link=False) as (data_from, data_to):
        data_to.objects = [name for name in data_from.objects if name in names]
    print('data_to.objects = ',data_to.objects)
    for obj in data_to.objects:
        if obj is not None:
#            scn.objects.link(obj)
#            bpy.data.collections['Collection'].objects.link(obj)
            bpy.data.collections['Collection'].objects.link(obj)
    grp = bpy.data.objects['00.00']
    arm = bpy.data.objects['1094395392']
    ob1 = bpy.data.objects['Object.000']
    ob2 = bpy.data.objects['Object.001']
    ob3 = bpy.data.objects['Object.002']
    if ('head' in ob1) == False:
        ob1['head'] = fullbase
#    group = scn.objects.link(grp)
#    armature = scn.objects.link(arm)          
    head = ob1 
    eyes = ob2
    teeth = ob3
    #parent objects together -- File>Append does this automagically?
#    group.makeParent([armature,head,eyes,teeth],0,1)
#    armature.makeParentDeform([head,eyes,teeth],0,0)    
    print('generic head loaded: %.4f sec.' % (time()-prev_time))
    prev_time = time()
    
    fv = faceverts(stream)
    print('faceverts read: %.4f sec.' % (time()-prev_time))
    prev_time = time()
    
    par1 = params(stream)
    print('params1 processed: %.4f sec.' % (time()-prev_time))
    prev_time = time()
    par2 = params(stream)
    print('params2 processed: %.4f sec.' % (time()-prev_time))
    prev_time = time()
    
    #process verts
    headmsh = head.data
    headvcount = len(headmsh.vertices)
    for v in headmsh.vertices:
        v.co[0] = fv[v.index].xpos
        v.co[1] = fv[v.index].ypos
        v.co[2] = fv[v.index].zpos
 #       v.normal[0] = fv[v.index].xnor
 #       v.normal[1] = fv[v.index].ynor
 #       v.normal[2] = fv[v.index].znor
    eyemsh = eyes.data
    eyevcount = len(eyemsh.vertices)
    for v in eyemsh.vertices:
        v.co[0] = fv[v.index + headvcount].xpos
        v.co[1] = fv[v.index + headvcount].ypos
        v.co[2] = fv[v.index + headvcount].zpos
#        v.normal[0] = fv[v.index + headvcount].xnor
#        v.normal[1] = fv[v.index + headvcount].ynor
#        v.normal[2] = fv[v.index + headvcount].znor
    teethmsh = teeth.data
    for v in teethmsh.vertices:
        v.co[0] = fv[v.index + headvcount + eyevcount].xpos
        v.co[1] = fv[v.index + headvcount + eyevcount].ypos
        v.co[2] = fv[v.index + headvcount + eyevcount].zpos
#        v.normal[0] = fv[v.index + headvcount + eyevcount].xnor
#        v.normal[1] = fv[v.index + headvcount + eyevcount].ynor
#        v.normal[2] = fv[v.index + headvcount + eyevcount].znor
    bpy.ops.object.shade_smooth()
    #Textures
    imagefile = header.texture1.strip().strip('\x00')
    testimage = os.path.join(filepath, imagefile)
    print('filepath = ',filepath)
    print('imagefile = ',imagefile)
    print('DEBUG: "%s"' % (testimage))
    if os.path.isfile(testimage):
        imageload = testimage
    else:
        print('imagedir = ',imagedir)
        print('imagefile = ',imagefile)
        imageload = get_image_path(imagedir, imagefile)
        print('imageload = ',imageload)
    if imageload:
        bpy.ops.image.open(filepath=imageload)
#        image = Image.Load(imageload)
    else:
        #imagefile = "%s"%imagefile

        image = Image.New(imagefile,32,32,24)
        print("Image", imagefile)
#    eyes texture = 'eyes.dds'
    imageload = get_image_path(imagedir,'eyes.dds')
    print('imageload = ',imageload)
    bpy.ops.image.open(filepath=imageload)
    #Material
#    mat = Material.New("Material")
    mats = bpy.data.materials
    mat = mats['Material']
    mat.specular_intensity = 0
#    mat.setMode('TexFace')
    #mat.setRGBCol(tmat.color_R/255.0,tmat.color_G/255.0,tmat.color_B/255.0)
    #mat.setAlpha(tmat.color_A/255.0)
#    myMap = Texture.MapTo['COL']           
      
    #Diffuse
    texname ="Diffuse"
#    t = Texture.New(texname)
#    t = bpy.data.textures.new('Diffuse',type='IMAGE')
#    t.setType('Image')
    print('images = ',bpy.data.images.keys())
# texture the head
    mat = head.data.materials[0]
    mat.use_nodes = True
    matnodes = mat.node_tree.nodes
    tex = matnodes.new('ShaderNodeTexImage')
    (x,y) = matnodes['Principled BSDF'].location
    tex.location = (x-300,y)
    tex.image = bpy.data.images[1]
    clr = matnodes['Principled BSDF'].inputs[0]
    mat.node_tree.links.new(clr,tex.outputs[0])
# texture the eyes
    mat = eyes.data.materials[0]
    mat.use_nodes = True
    matnodes = mat.node_tree.nodes
    tex = matnodes.new('ShaderNodeTexImage')
    (x,y) = matnodes['Principled BSDF'].location
    tex.location = (x-300,y)
    tex.image = bpy.data.images[0]
    clr = matnodes['Principled BSDF'].inputs[0]
    mat.node_tree.links.new(clr,tex.outputs[0])
# texture the eyes
    mat = teeth.data.materials[0]
    mat.use_nodes = True
    matnodes = mat.node_tree.nodes
    tex = matnodes.new('ShaderNodeTexImage')
    (x,y) = matnodes['Principled BSDF'].location
    tex.location = (x-300,y)
    tex.image = bpy.data.images[0]
    clr = matnodes['Principled BSDF'].inputs[0]
    mat.node_tree.links.new(clr,tex.outputs[0])            
#    mat.setTexture(0, t, Blender.Texture.TexCo.UV, myMap)
#    mat.active_texture = t
    print('texture/image loaded: %.4f sec.' % (time()-prev_time))
    prev_time = time()    
    print('\thd file loaded TOTAL: %.4f sec.' % (time()-start_time))

#***********************************************
# register callback
#***********************************************
#def my_callback(filename):
#    load_hd(filename)
#filename = 'C:\\TMBlender\\Ble279\\scripts\\head_m_white_man2.hd'
#load_hd(filename)
#Blender.Window.FileSelector(my_callback, "Import .hd", '*.hd')
def read_hd(context,filepath):
    load_hd(filepath)
    return {'FINISHED'}
class TheMoviesHeadImporter(Operator, ImportHelper):
    """This appears in the tooltip of the operator and in the generated docs"""
    bl_idname = "object.head_import"  # important since its how bpy.ops.import_test.some_data is constructed
    bl_label = "Import TM HD"

    # ImportHelper mixin class uses this
    filename_ext = ".hd"

    filter_glob = StringProperty(
        default="*.hd",
        options={'HIDDEN'},
        maxlen=255,  # Max internal buffer length, longer would be clamped.
    )

    def execute(self, context):
        return read_hd(context,self.filepath)

# Only needed if you want to add into a dynamic menu
def menu_func_import(self, context):
    self.layout.operator(TheMoviesHeadImporter.bl_idname, text="Import The Movies Head (*.hd)")

def register():
    bpy.utils.register_class(TheMoviesHeadImporter)
    bpy.types.TOPBAR_MT_file_import.append(menu_func_import)
#    bpy.types.INFO_MT_file_import.append(menu_func_import)    
def unregister():
#    bpy.utils.unregister_class(TheMoviesHeadImporter)
    bpy.types.TOPBAR_MT_file_import.remove(menu_func_import)
#    bpy.types.INFO_MT_file_import.append(menu_func_import)

#def hdin():
#    bpy.ops.object.head_import('INVOKE_DEFAULT')
if __name__ == "__main__":
    register()
#    bpy.ops.object.head_import('INVOKE_DEFAULT')
#    register()
    
    # test call
#    hdin()
